PAGE 58,132
TITLE MAKETHRD.ASM - Example VxD for Win95

; ==========================================================
; THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF 
; ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO 
; THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A 
; PARTICULAR PURPOSE.
;
; (C)Copyright 1995, Microsoft Corporation. All rights reserved.
;
; MAKETHRD.VXD - Sample VxD which services requests from ring
; 3 to create and destroy a ring 0 thread. The threads it
; creates calls ring 3 to put up a message "One thread created".
; and waits on a Win32 event.
;
; When the user dismisses the dialog box, the ring0 thread
; is woken from the event and terminates.
;
; ==========================================================

.386P

option	readonly			; Make it pure
option	segment:flat			; Use flat model
option	offset:flat
use32	textequ <flat public>		; Make use32 an alias for flat

.xlist
INCLUDE VMM.INC
INCLUDE VWIN32.INC
.list

; ==========================================================
; Device function and error codes
; ==========================================================
METHOD_NEITHER	EQU 3
FILE_ANY_ACCESS	EQU 0

CTL_CODE	MACRO CodeName, DeviceType, Function, Method, Access
    &CodeName EQU (((DeviceType) SHL 16) + ((Access) SHL 14) + ((Function) SHL 2) + (Method))
ENDM

DEVICE_WIN95VXD	EQU 32768

; get version code, ring 0 limited
MAKETHRD_GetVer		EQU 0

; The following codes can be used from ring 3 without compromising
; NT compatibility.

; Get version from ring 3
CTL_CODE MAKETHRD_GetVerR3, DEVICE_WIN95VXD, 2048, METHOD_NEITHER, FILE_ANY_ACCESS
; make a ring 0 thread
CTL_CODE MAKETHRD_MakeThread, DEVICE_WIN95VXD, 2049, METHOD_NEITHER, FILE_ANY_ACCESS
	
MAKETHRD_Ver_Major	EQU 4
MAKETHRD_Ver_Minor	EQU 0

; we don't have these Win32 codes in an include file
ERROR_NOT_ENOUGH_MEMORY	EQU 50
ERROR_NOT_SUPPORTED	EQU 50
ERROR_INVALID_PARAMETER	EQU 87
INFINITE		EQU -1

; ==========================================================
; Structure used here
; ==========================================================
THREAD_START_STRUC	STRUC
  hRing0Event		dd 0		; VxDHandle to Win32 event
  hRing3Event		dd 0		; parameter for ring 3 function
  hVxD			dd 0		; VxD handle to pass back to ring 3
  hRing0Thread		dd 0		; ring 3 handle to ring 0 thread
  pfnRing3APC		dd 0		; function to call at ring 3
  thcbUserAPC		dd 0		; thread to queue user APC to
THREAD_START_STRUC	ENDS


; ==========================================================
; Virtual device declaration - sets up the DDB
; ==========================================================
; the following equate makes the VXD dynamically loadable and name based.
MAKETHRD_DYNAMIC	EQU TRUE
MAKETHRD_NAME_BASED	EQU TRUE
MAKETHRD_Device_ID	EQU UNDEFINED_DEVICE_ID
MAKETHRD_Version	EQU (MAKETHRD_Ver_Major SHL 8)+MAKETHRD_Ver_Minor

Declare_Virtual_Device	MAKETHRD, MAKETHRD_Ver_Major, MAKETHRD_Ver_Minor,\
			MAKETHRD_Control, MAKETHRD_Device_ID,\
			UNDEFINED_INIT_ORDER, ,


; ==========================================================
; Useful little macro
; ==========================================================
Debug_Out MACRO StrLiteral
    LOCAL StrLabel
    IFDEF DEBUG
	VxD_LOCKED_DATA_SEG
	  StrLabel db StrLiteral,13,10,0
	VxD_LOCKED_DATA_ENDS
	pushfd			;; save flags
	pushad			;; and regs
	mov	esi,OFFSET StrLabel ;; point to string
	VMMCall	Out_Debug_String ;; write it
	popad			;; restore regs
	popfd			;; and flags
    ENDIF
ENDM

; ==========================================================
; Global variable area
; ==========================================================
VxD_PAGEABLE_DATA_SEG
VxD_PAGEABLE_DATA_ENDS

; ==========================================================
; The only locked code in our VxD is the control dispatcher.
; ==========================================================
VxD_LOCKED_CODE_SEG

; ==========================================================
; We only handle the DeviceIOControl control message. We
; really don't need to handle Device_Init since we have
; nothing to initialize.
; ==========================================================
BeginProc MAKETHRD_Control

	Control_Dispatch W32_DEVICEIOCONTROL, MAKETHRD_DIOC_Handler
	clc
	ret

EndProc MAKETHRD_Control

VxD_LOCKED_CODE_ENDS


;==========================================================
; Our pageable code starts here.
;==========================================================
VxD_PAGEABLE_CODE_SEG

;==========================================================
; MAKETHRD_StoreReturnCode - since there isn't a way to
;	return a DWORD besides error code to DeviceIOControl
;	calls, this function stores a return code in eax
;	in the output buffer if possible. If not, it returns
;	carry for error.
;
; Input:
;	esi - pointer to DIOCPARAMETERS
;	eax - code to store in the output buffer
;
; Output:
;	CY flag set if the buffer was not 4 bytes in size
;	NC if the code was set. Also stores 4 in bytes returned.
;==========================================================
BeginProc MAKETHRD_StoreReturnCode, PUBLIC, ccall

	cmp	[esi].cbOutBuffer,4	; make sure we have room
	jb	MSRCReturn		; if not, return failure

	mov	ecx,[esi].lpvOutBuffer	; point to buffer and store code
	mov	dword ptr [ecx],eax	; done
	mov	ecx,[esi].lpcbBytesReturned ; get pointer to bytes returned
	jecxz	MSRCReturn		; if no pointer, return

	mov	dword ptr [ecx],4	; store number of bytes returned

      MSRCReturn:
	ret				; return

EndProc MAKETHRD_StoreReturnCode

;==========================================================
; DeviceIoControl API HANDLER
; To use this, the Win32 EXE will call:
;
;	hFile = CreateFile(\\.\MAKETHRD.VXD, ...)
;
; Then, DeviceIOControl(hFIle, dwIoControlCode, lpvInBuffer, ...)
;
; Input:
;	esi - pointer to DIOCPARAMETERS from VWIN32.INC
;
; Output:
;	function 0, get version
;	function 1, thread create
;==========================================================
BeginProc MAKETHRD_DIOC_Handler, PUBLIC

	mov	ecx,[esi].dwIoControlCode	; get the code

	Debug_Out "MAKETHRD_DIOC_Handler -- function #ecx"

	jecxz	MAKETHRD_Get_Version

	cmp	ecx,MAKETHRD_GetVerR3		; is it ring 3 get version?
	jz	MAKETHRD_Get_Version		; if so, branch
	cmp	ecx,MAKETHRD_MakeThread		; should we make a thread?
	jz	MAKETHRD_Make_Thread		; if so, do it

	mov	eax,ERROR_NOT_SUPPORTED		; set error
	ret					; and return

EndProc MAKETHRD_DIOC_Handler

;==========================================================
; MAKETHRD_DIOC_Get_Version - always returns OK.
;
; Input:
;	esi - pointer to DIOCPARAMETERS from VWIN32.INC
;
; Output:
;	eax = 0
;	The convention is to return version information in
;	ecx, but if you do, it will be thrown away.
;	version information is returned in the return buffer
;	if the return buffer pointer is <> NULL, and the count
;	field is 2.
;==========================================================
BeginProc MAKETHRD_Get_Version, PUBLIC

	mov	eax,(MAKETHRD_Ver_Major SHL 8) + MAKETHRD_Ver_Minor
	cCall	MAKETHRD_StoreReturnCode ; store as return code
	xor	eax,eax			; say we succeeded
	ret				; and return

EndProc MAKETHRD_Get_Version

;==========================================================
; MAKETHRD_Make_Thread - our service API which creates a
;	ring 0 thread. If successful, a ring 3 thread handle
;	is returned which can be used to set/get priorities.
;	This handle should be closed at ring 3.
;
; Input:
;	esi - pointer to DIOCPARAMETERS from VWIN32.INC
;
; Output:
;	Thread created and ring 3 thread handle returned
;	in output buffer or error returned.
;==========================================================
BeginProc MAKETHRD_Make_Thread, PUBLIC, ESP
	LocalVar DIOCParamPtr,DWORD

	EnterProc			; setup stack frame

	Debug_Out "MAKETHRD_Make_Thread"

	mov	DIOCParamPtr,esi	; save parameter pointer

	; make sure we've been passed the right size buffer
	mov	ecx,[esi].cbInBuffer	; get count of bytes
	cmp	ecx,(SIZEOF THREAD_START_STRUC)
	jz	MMTSetupStruc		; if we have, setup thread start struc

      MMTRetInvalParm:
	mov	eax,ERROR_INVALID_PARAMETER ; not valid
	jmp	MMTReturn		; return

      MMTSetupStruc:
	mov	esi,[esi].lpvInBuffer	; get pointer to buffer
	test	esi,esi			; is it NULL
	jz	MMTRetInvalParm		; if no pointer, return invalid

	; allocate the thread startup structure
	VMMCall _HeapAllocate, <SIZEOF THREAD_START_STRUC, HeapZeroInit>
	or	eax,eax			; make sure we got the memory
	jnz	MMTMoveParms		; if so, continue

      MMTNotEnoMem:
	mov	eax,ERROR_NOT_ENOUGH_MEMORY ; otherwise, return error
	jmp	MMTReturn		; return

      MMTMoveParms:
	mov	edi,eax			; get memory in destination
	mov	edx,edi			; and parameter to create thread
	mov	ecx,(SIZEOF THREAD_START_STRUC)
	cld				; forward direction
	rep	movsb			; copy the structure to global mem

	VMMCall	Get_Cur_Thread_Handle	; get current thread handle in edi
	mov	[edx].thcbUserAPC,edi	; store thread handle for APC

	mov	edi,edx			; refresh edi to store output buffer

	; ecx is 0 for default stack size
	; edx points to parameter block
	mov	esi,ecx			; FALSE for synchronous create
	mov	ebx,OFFSET MAKETHRD_ThreadStartup ; startup function
	VxDCall	_VWin32_CreateRing0Thread ; make the thread
	test	eax,eax			; if not 0,
	jnz	MMTHaveThread		; we have a thread

	; if we couldn't create the thread, we need to close
	; the ring 0 event handle we were passed. that can
	; only be done at ring 0. we also free the memory and 
	; return
	mov	eax,[edi].hRing0Event	; get event handle
	VxDCall	_VWin32_CloseVxDHandle	; and close it

	VMMCall	_HeapFree, <edi,0>	; free the parameter block
	jmp	MMTNotEnoMem		; just return not enough memory

      MMTHaveThread:
	mov	[edi].hRing0Thread,eax	; store thread handle in struct
	mov	esi,DIOCParamPtr	; get parameter block ptr
	cCall	MAKETHRD_StoreReturnCode ; store the return code
	xor	eax,eax			; return OK

      MMTReturn:
	LeaveProc			; clear stack params
	Return				; and return

EndProc MAKETHRD_Make_Thread

;==========================================================
; MAKETHRD_ThreadStartup - this is the function in which
;	a new ring 0 thread executes.
;
; Input:
;	one parameter which points to a memory block containing
;	an event handle, a ring 3 APC function, and the ring 0
;	thread handle to which we'll queue the APC.
;
; Output:
;	we queue the APC, block on the event, when it is signaled,
;	we terminate by returning.
;==========================================================
BeginProc MAKETHRD_ThreadStartup, PUBLIC, ccall, ESP
	ArgVar dwThreadParam,DWORD	; our thread parameter

	EnterProc			; start function

	Debug_Out "MAKETHRD_ThreadStartup"

	mov	edi,dwThreadParam	; point to thread parameter block

	; queue an APC to the ring 3 app thread
	VxDCall	_VWin32_QueueUserAPC, <[edi].pfnRing3APC,edi,[edi].thcbUserAPC>

	mov	ebx,[edi].hRing0Event	; get event handle
	mov	edx,INFINITE		; wait forever
	xor	eax,eax			; not alertable
	VxDCall	_VWin32_WaitSingleObject ; wait for the event

	mov	eax,[edi].hRing0Event	; get event handle to close
	VxDCall	_VWin32_CloseVxDHandle	; close it

	VMMCall	_HeapFree, <edi,0>	; free the structure

	LeaveProc			; end of variable usage
	Return				; return to terminate thread

EndProc MAKETHRD_ThreadStartup


VxD_PAGEABLE_CODE_ENDS
END

